@@ -34,6 +34,7 @@
 #include "qgsskyboxsettings.h"
 #include "qgsshadowsettings.h"
 #include "qgscameracontroller.h"
+#include "qgstemporalrangeobject.h"
 
 class QgsMapLayer;
 class QgsRasterLayer;
@@ -58,7 +59,7 @@ class _3D_EXPORT Qgs3DMapSettings : public QObject, public QgsTemporalRangeObjec
   public:
 
     //! Constructor for Qgs3DMapSettings
-    Qgs3DMapSettings() = default;
+    Qgs3DMapSettings();
     //! Copy constructor
     Qgs3DMapSettings( const Qgs3DMapSettings &other );
     ~Qgs3DMapSettings() override;
@@ -163,11 +164,8 @@ class _3D_EXPORT Qgs3DMapSettings : public QObject, public QgsTemporalRangeObjec
      *
      * This setting dictates which layers are to be rendered using their 3D rendering configuration, if available.
      *
-     * \note Layers which are rendered as part of the map terrain are specified via \a setTerrainLayers().
-     *
      * \see layers()
      * \see layersChanged()
-     * \see setTerrainLayers()
      */
     void setLayers( const QList<QgsMapLayer *> &layers );
 
@@ -176,11 +174,8 @@ class _3D_EXPORT Qgs3DMapSettings : public QObject, public QgsTemporalRangeObjec
      *
      * This setting dictates which layers are to be rendered using their 3D rendering configuration, if available.
      *
-     * \note Layers which are rendered as part of the map terrain are retrieved via \a terrainLayers().
-     *
      * \see setLayers()
      * \see layersChanged()
-     * \see terrainLayers()
      */
     QList<QgsMapLayer *> layers() const;
 
@@ -188,34 +183,6 @@ class _3D_EXPORT Qgs3DMapSettings : public QObject, public QgsTemporalRangeObjec
     // terrain related config
     //
 
-    /**
-     * Sets the list of 2d map \a layers to be rendered in the terrain.
-     *
-     * \note Layers which are rendered as 3D layers as part of the scene are specified via \a setLayers().
-     *
-     * \note If terrainMapTheme() is set, it has a priority over the list of layers specified here.
-     *
-     * \see terrainLayers()
-     * \see terrainLayersChanged()
-     * \see setLayers()
-     * \since QGIS 3.16
-     */
-    void setTerrainLayers( const QList<QgsMapLayer *> &layers );
-
-    /**
-     * Returns the list of map layers to be rendered as a texture of the terrain.
-     *
-     * \note Layers which are rendered as 3D layers as part of the scene are retrieved via \a layers().
-     *
-     * \note If terrainMapTheme() is set, it has a priority over the list of layers returned here.
-     *
-     * \see setTerrainLayers()
-     * \see terrainLayersChanged()
-     * \see layers()
-     * \since QGIS 3.16
-     */
-    QList<QgsMapLayer *> terrainLayers() const;
-
     /**
      * Sets vertical scale (exaggeration) of terrain
      * (1 = true scale, > 1 = hills get more pronounced)
@@ -283,7 +250,12 @@ class _3D_EXPORT Qgs3DMapSettings : public QObject, public QgsTemporalRangeObjec
      */
     void setTerrainGenerator( QgsTerrainGenerator *gen SIP_TRANSFER ) SIP_SKIP;
     //! Returns terrain generator. It takes care of producing terrain tiles from the input data.
-    QgsTerrainGenerator *terrainGenerator() const SIP_SKIP { return mTerrainGenerator.get(); }
+    QgsTerrainGenerator *terrainGenerator() const SIP_SKIP
+    {
+      if ( mTerrainRenderingEnabled )
+        return mTerrainGenerator.get();
+      return nullptr;
+    }
 
     /**
      * Sets whether terrain shading is enabled.
@@ -359,6 +331,18 @@ class _3D_EXPORT Qgs3DMapSettings : public QObject, public QgsTemporalRangeObjec
      */
     bool showCameraViewCenter() const { return mShowCameraViewCenter; }
 
+    /**
+     * Sets whether to show camera's rotation center as a sphere (for debugging)
+     * \since QGIS 3.24
+     */
+    void setShowCameraRotationCenter( bool enabled );
+
+    /**
+     * Returns whether to show camera's rotation center as a sphere (for debugging)
+     * \since QGIS 3.24
+     */
+    bool showCameraRotationCenter() const { return mShowCameraRotationCenter; }
+
     /**
      * Sets whether to show light source origins as a sphere (for debugging)
      * \since QGIS 3.16
@@ -571,7 +555,45 @@ class _3D_EXPORT Qgs3DMapSettings : public QObject, public QgsTemporalRangeObjec
      */
     void setIsFpsCounterEnabled( bool fpsCounterEnabled );
 
+    /**
+     * Returns whether the 2D terrain surface will be rendered.
+     * \see setTerrainRenderingEnabled()
+     * \since QGIS 3.22
+     */
+    bool terrainRenderingEnabled() { return mTerrainRenderingEnabled; }
+
+    /**
+     * Sets whether the 2D terrain surface will be rendered in.
+     * \see terrainRenderingEnabled()
+     * \since QGIS 3.22
+     */
+    void setTerrainRenderingEnabled( bool terrainRenderingEnabled );
+
+    /**
+     * Returns the renderer usage
+     *
+     * \see rendererUsage()
+     * \since QGIS 3.24
+     */
+    Qgis::RendererUsage rendererUsage() const;
+
+    /**
+     * Sets the renderer usage
+     *
+     * \see rendererUsage()
+     * \since QGIS 3.24
+     */
+    void setRendererUsage( Qgis::RendererUsage rendererUsage );
+
   signals:
+
+    /**
+     * Emitted when one of the configuration settings has changed
+     *
+     * \since QGIS 3.24
+     */
+    void settingsChanged();
+
     //! Emitted when the background color has changed
     void backgroundColorChanged();
     //! Emitted when the selection color has changed
@@ -582,21 +604,9 @@ class _3D_EXPORT Qgs3DMapSettings : public QObject, public QgsTemporalRangeObjec
      *
      * \see setLayers()
      * \see layers()
-     * \see terrainLayersChanged()
      */
     void layersChanged();
 
-    /**
-     * Emitted when the list of map layers for terrain texture has changed.
-     *
-     * \see terrainLayers()
-     * \see setTerrainLayers()
-     * \see layersChanged()
-     *
-     * \since QGIS 3.16
-     */
-    void terrainLayersChanged();
-
     //! Emitted when the terrain generator has changed
     void terrainGeneratorChanged();
     //! Emitted when the vertical scale of the terrain has changed
@@ -643,6 +653,12 @@ class _3D_EXPORT Qgs3DMapSettings : public QObject, public QgsTemporalRangeObjec
      */
     void showCameraViewCenterChanged();
 
+    /**
+     * Emitted when the flag whether camera's rotation center is shown has changed
+     * \since QGIS 3.24
+     */
+    void showCameraRotationCenterChanged();
+
     /**
      * Emitted when the flag whether light source origins are shown has changed.
      * \since QGIS 3.15
@@ -732,7 +748,6 @@ class _3D_EXPORT Qgs3DMapSettings : public QObject, public QgsTemporalRangeObjec
 
     /**
      * Emitted when the FPS counter is enabled or disabled
-     *
      * \since QGIS 3.18
      */
     void fpsCounterEnabledChanged( bool fpsCounterEnabled );
@@ -742,6 +757,10 @@ class _3D_EXPORT Qgs3DMapSettings : public QObject, public QgsTemporalRangeObjec
     Qgs3DMapSettings &operator=( const Qgs3DMapSettings & );
 #endif
 
+  private:
+    //! Connects the various changed signals of this widget to the settingsChanged signal
+    void connectChangedSignalsToSettingsChanged();
+
   private:
     //! Offset in map CRS coordinates at which our 3D world has origin (0,0,0)
     QgsVector3D mOrigin;
@@ -760,6 +779,7 @@ class _3D_EXPORT Qgs3DMapSettings : public QObject, public QgsTemporalRangeObjec
     bool mShowTerrainBoundingBoxes = false;  //!< Whether to show bounding boxes of entities - useful for debugging
     bool mShowTerrainTileInfo = false;  //!< Whether to draw extra information about terrain tiles to the textures - useful for debugging
     bool mShowCameraViewCenter = false;  //!< Whether to show camera view center as a sphere - useful for debugging
+    bool mShowCameraRotationCenter = false; //!< Whether to show camera rotation center as a sphere - useful for debugging
     bool mShowLightSources = false; //!< Whether to show the origin of light sources
     bool mShowLabels = false; //!< Whether to display labels on terrain tiles
     QList<QgsPointLightSettings> mPointLights;  //!< List of point lights defined for the scene
@@ -769,7 +789,6 @@ class _3D_EXPORT Qgs3DMapSettings : public QObject, public QgsTemporalRangeObjec
     QgsCameraController::NavigationMode mCameraNavigationMode = QgsCameraController::NavigationMode::TerrainBasedNavigation;
     double mCameraMovementSpeed = 5.0;
     QList<QgsMapLayerRef> mLayers;   //!< Layers to be rendered
-    QList<QgsMapLayerRef> mTerrainLayers;   //!< Terrain layers to be rendered
     QList<QgsAbstract3DRenderer *> mRenderers;  //!< Extra stuff to render as 3D object
     //! Coordinate transform context
     QgsCoordinateTransformContext mTransformContext;
@@ -793,6 +812,10 @@ class _3D_EXPORT Qgs3DMapSettings : public QObject, public QgsTemporalRangeObjec
     bool mDebugDepthMapEnabled = false;
     Qt::Corner mDebugDepthMapCorner = Qt::Corner::TopRightCorner;
     double mDebugDepthMapSize = 0.2;
+
+    bool mTerrainRenderingEnabled = true;
+
+    Qgis::RendererUsage mRendererUsage;
 };
 
 
