@@ -32,6 +32,7 @@
 
 //qgis includes
 #include "qgis_sip.h"
+#include "qgsconfig.h"
 #include "qgsunittypes.h"
 #include "qgsrectangle.h"
 #include "qgssqliteutils.h"
@@ -39,15 +40,20 @@
 class QDomNode;
 class QDomDocument;
 class QgsCoordinateReferenceSystemPrivate;
+class QgsDatumEnsemble;
+class QgsProjectionFactors;
+class QgsProjOperation;
 
-#if PROJ_VERSION_MAJOR>=6
 #ifndef SIP_RUN
 struct PJconsts;
 typedef struct PJconsts PJ;
 
+#if PROJ_VERSION_MAJOR>=8
+struct pj_ctx;
+typedef struct pj_ctx PJ_CONTEXT;
+#else
 struct projCtx_t;
 typedef struct projCtx_t PJ_CONTEXT;
-
 #endif
 #endif
 
@@ -65,7 +71,7 @@ typedef void ( *CUSTOM_CRS_VALIDATION )( QgsCoordinateReferenceSystem & ) SIP_SK
 
 /**
  * \ingroup core
- * This class represents a coordinate reference system (CRS).
+ * \brief This class represents a coordinate reference system (CRS).
  *
  * Coordinate reference system object defines a specific map projection, as well as transformations
  * between different coordinate reference systems. There are various ways how a CRS can be defined:
@@ -103,7 +109,7 @@ typedef void ( *CUSTOM_CRS_VALIDATION )( QgsCoordinateReferenceSystem & ) SIP_SK
  *
  * This will produce the following output:
  *
- * \code
+ * \code{.unparsed}
  * CRS Description: OSGB 1936 / British National Grid
  * CRS PROJ text: +proj=tmerc +lat_0=49 +lon_0=-2 +k=0.9996012717 +x_0=400000 +y_0=-100000 [output trimmed]
  * \endcode
@@ -132,7 +138,7 @@ typedef void ( *CUSTOM_CRS_VALIDATION )( QgsCoordinateReferenceSystem & ) SIP_SK
  *   format: `+param1=value1 +param2=value2 [...]`. This is the format natively used by the
  *   underlying proj library. For example, the definition of WGS84 looks like this:
  *
- *   \code
+ *   \code{.unparsed}
  *   +proj=longlat +datum=WGS84 +no_defs
  *   \endcode
  *
@@ -142,7 +148,7 @@ typedef void ( *CUSTOM_CRS_VALIDATION )( QgsCoordinateReferenceSystem & ) SIP_SK
  * - Well-known text (WKT): Defined by Open Geospatial Consortium (OGC), this is another common
  *   format to define CRS. For WGS84 the OGC WKT definition is the following:
  *
- *   \code
+ *   \code{.unparsed}
  *       GEOGCS["WGS 84",
  *              DATUM["WGS_1984",
  *                SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],
@@ -208,6 +214,8 @@ class CORE_EXPORT QgsCoordinateReferenceSystem
 
     Q_PROPERTY( QgsUnitTypes::DistanceUnit mapUnits READ mapUnits )
     Q_PROPERTY( bool isGeographic READ isGeographic )
+    Q_PROPERTY( QString authid READ authid )
+    Q_PROPERTY( QString description READ description )
 
   public:
 
@@ -219,13 +227,6 @@ class CORE_EXPORT QgsCoordinateReferenceSystem
       EpsgCrsId       //!< EPSG code
     };
 
-    //! Projection definition formats
-    enum Format
-    {
-      FormatWkt = 0, //!< WKT format (always recommended over proj string format)
-      FormatProj, //!< Proj string format
-    };
-
     //! Constructs an invalid CRS object
     QgsCoordinateReferenceSystem();
 
@@ -728,6 +729,112 @@ class CORE_EXPORT QgsCoordinateReferenceSystem
      */
     bool isGeographic() const;
 
+    /**
+     * Returns TRUE if the CRS is a dynamic CRS.
+     *
+     * A dynamic CRS relies on a dynamic datum, that is a datum that is not
+     * plate-fixed.
+     *
+     * \since QGIS 3.20
+     */
+    bool isDynamic() const;
+
+    /**
+     * Attempts to retrieve datum ensemble details from the CRS.
+     *
+     * If the CRS does not use a datum ensemble then an invalid QgsDatumEnsemble will
+     * be returned.
+     *
+     * \warning This method requires PROJ 8.0 or later
+     *
+     * \throws QgsNotSupportedException on QGIS builds based on PROJ 7 or earlier.
+     *
+     * \since QGIS 3.20
+     */
+    QgsDatumEnsemble datumEnsemble() const SIP_THROW( QgsNotSupportedException );
+
+    /**
+     * Attempts to retrieve the name of the celestial body associated with the CRS (e.g. "Earth").
+     *
+     * \warning This method requires PROJ 8.1 or later
+     *
+     * \throws QgsNotSupportedException on QGIS builds based on PROJ 8.0 or earlier.
+     *
+     * \since QGIS 3.20
+     */
+    QString celestialBodyName() const SIP_THROW( QgsNotSupportedException );
+
+    /**
+     * Sets the coordinate \a epoch, as a decimal year.
+     *
+     * In a dynamic CRS (see isDynamic()), coordinates of a point on the surface of the Earth may
+     * change with time. To be unambiguous the coordinates must always be qualified
+     * with the epoch at which they are valid. The coordinate epoch is not necessarily
+     * the epoch at which the observation was collected.
+     *
+     * Pedantically the coordinate epoch of an observation belongs to the
+     * observation, and not to the CRS, however it is often more practical to
+     * bind it to the CRS. The coordinate epoch should be specified for dynamic
+     * CRS (see isDynamic()).
+     *
+     * \param epoch Coordinate epoch as decimal year (e.g. 2021.3)
+     *
+     * \warning The QgsCoordinateTransform class can perform time-dependent transformations
+     * between a static and dynamic CRS based on either the source or destination CRS coordinate epoch,
+     * however dynamic CRS to dynamic CRS transformations are not currently supported.
+     *
+     * \see coordinateEpoch()
+     *
+     * \since QGIS 3.20
+     */
+    void setCoordinateEpoch( double epoch );
+
+    /**
+     * Returns the coordinate epoch, as a decimal year.
+     *
+     * In a dynamic CRS, coordinates of a point on the surface of the Earth may
+     * change with time. To be unambiguous the coordinates must always be qualified
+     * with the epoch at which they are valid. The coordinate epoch is not necessarily
+     * the epoch at which the observation was collected.
+     *
+     * Pedantically the coordinate epoch of an observation belongs to the
+     * observation, and not to the CRS, however it is often more practical to
+     * bind it to the CRS. The coordinate epoch should be specified for dynamic
+     * CRS (see isDynamic()).
+     *
+     * \warning The QgsCoordinateTransform class can perform time-dependent transformations
+     * between a static and dynamic CRS based on either the source or destination CRS coordinate epoch,
+     * however dynamic CRS to dynamic CRS transformations are not currently supported.
+     *
+     * \returns Coordinate epoch as decimal year (e.g. 2021.3), or NaN if not set, or relevant.
+     *
+     * \see setCoordinateEpoch()
+     *
+     * \since QGIS 3.20
+     */
+    double coordinateEpoch() const;
+
+    /**
+     * Calculate various cartographic properties, such as scale factors, angular distortion and meridian convergence for
+     * the CRS at the given geodetic \a point (in geographic coordinates).
+     *
+     * Depending on the underlying projection values will be calculated either numerically (default) or analytically.
+     * The function also calculates the partial derivatives of the given coordinate.
+     *
+     * \note Internally uses the proj library proj_factors API to calculate the factors.
+     *
+     * \since QGIS 3.20
+     */
+    QgsProjectionFactors factors( const QgsPoint &point ) const;
+
+    /**
+     * Returns information about the PROJ operation associated with the coordinate reference system, for example
+     * the projection method used by the CRS.
+     *
+     * \since QGIS 3.20
+     */
+    QgsProjOperation operation() const;
+
     /**
      * Returns whether axis is inverted (e.g., for WMS 1.3) for the CRS.
      * \returns TRUE if CRS axis is inverted
@@ -751,6 +858,24 @@ class CORE_EXPORT QgsCoordinateReferenceSystem
 
     // Mutators -----------------------------------
 
+    /**
+     * Updates the definition and parameters of the coordinate reference system to their
+     * latest values.
+     *
+     * This only has an effect if the CRS is a user defined custom CRS, and the definition
+     * of that custom CRS has changed. In this case the parameters of the object (such as the
+     * proj and WKT string definitions, and other related properties) will be updated to
+     * reflect the current definition of the custom CRS.
+     *
+     * Any objects which store CRS objects should connect to the QgsApplication::coordinateReferenceSystemRegistry()'s
+     * QgsCoordinateReferenceSystemRegistry::userCrsChanged() signal and call this method
+     * on their stored CRS objects whenever the signal is emitted in order to update these
+     * CRSes to their new definitions.
+     *
+     * \since QGIS 3.18
+     */
+    void updateDefinition();
+
     /**
      * Set user hint for validation
      */
@@ -770,7 +895,7 @@ class CORE_EXPORT QgsCoordinateReferenceSystem
     static int syncDatabase();
 
     /**
-     * Saves the CRS as a custom ("USER") CRS.
+     * Saves the CRS as a new custom ("USER") CRS.
      *
      * Returns the new CRS srsid(), or -1 if the CRS could not be saved.
      *
@@ -779,8 +904,43 @@ class CORE_EXPORT QgsCoordinateReferenceSystem
      *
      * \warning Not all CRS definitions can be represented as a Proj string, so
      * take care when using the FormatProj option.
+     *
+     * \note Since QGIS 3.18, internally this calls QgsCoordinateReferenceSystemRegistry::addUserCrs().
      */
-    long saveAsUserCrs( const QString &name, Format nativeFormat = FormatWkt );
+    long saveAsUserCrs( const QString &name, Qgis::CrsDefinitionFormat nativeFormat = Qgis::CrsDefinitionFormat::Wkt );
+
+    /**
+     * Sets the native \a format for the CRS definition.
+     *
+     * \note This has no effect on the underlying definition of the CRS, rather it controls what format
+     * to use when displaying the CRS's definition to users.
+     *
+     * \see nativeFormat()
+     * \since QGIS 3.24
+     */
+    void setNativeFormat( Qgis::CrsDefinitionFormat format );
+
+    /**
+     * Returns the native format for the CRS definition.
+     *
+     * \note This has no effect on the underlying definition of the CRS, rather it controls what format
+     * to use when displaying the CRS's definition to users.
+     *
+     * \see setNativeFormat()
+     * \since QGIS 3.24
+     */
+    Qgis::CrsDefinitionFormat nativeFormat() const;
+
+    /**
+     * Returns the geographic CRS associated with this CRS object.
+     *
+     * May return an invalid CRS if the geographic CRS could not be determined.
+     *
+     * \note This method will always return a longitude, latitude ordered CRS.
+     *
+     * \since QGIS 3.24
+     */
+    QgsCoordinateReferenceSystem toGeographicCrs() const;
 
     //! Returns auth id of related geographic CRS
     QString geographicCrsAuthId() const;
@@ -788,13 +948,14 @@ class CORE_EXPORT QgsCoordinateReferenceSystem
 #ifdef SIP_RUN
     SIP_PYOBJECT __repr__();
     % MethodCode
-    QString str = QStringLiteral( "<QgsCoordinateReferenceSystem: %1>" ).arg( sipCpp->authid() );
+    const QString str = sipCpp->isValid() ? QStringLiteral( "<QgsCoordinateReferenceSystem: %1%2>" ).arg( !sipCpp->authid().isEmpty() ? sipCpp->authid() : sipCpp->toWkt( QgsCoordinateReferenceSystem::WKT_PREFERRED ),
+                        std::isfinite( sipCpp->coordinateEpoch() ) ? QStringLiteral( " @ %1" ).arg( sipCpp->coordinateEpoch() ) : QString() )
+                        : QStringLiteral( "<QgsCoordinateReferenceSystem: invalid>" );
     sipRes = PyUnicode_FromString( str.toUtf8().constData() );
     % End
 #endif
 
 #ifndef SIP_RUN
-#if PROJ_VERSION_MAJOR>=6
 
     /**
      * Returns the underlying PROJ PJ object corresponding to the CRS, or NULLPTR
@@ -806,7 +967,28 @@ class CORE_EXPORT QgsCoordinateReferenceSystem
      * \since QGIS 3.8
      */
     PJ *projObject() const;
-#endif
+
+    /**
+     * Constructs a QgsCoordinateReferenceSystem from a PROJ PJ object.
+     *
+     * The \a object must correspond to a PROJ CRS object.
+     *
+     * Ownership of \a object is not transferred.
+     *
+     * \note Not available in Python bindings
+     * \since QGIS 3.24
+     */
+    static QgsCoordinateReferenceSystem fromProjObject( PJ *object );
+
+    /**
+     * Sets this CRS by passing it a PROJ PJ \a object, corresponding to a PROJ CRS object.
+     *
+     * Ownership of \a object is not transferred.
+     *
+     * \note Not available in Python bindings
+     * \since QGIS 3.24
+     */
+    bool createFromProjObject( PJ *object );
 #endif
 
     /**
@@ -875,7 +1057,7 @@ class CORE_EXPORT QgsCoordinateReferenceSystem
     /**
      * Set the WKT string
      */
-    bool setWktString( const QString &wkt, bool allowProjFallback = true );
+    bool setWktString( const QString &wkt );
 
     /**
      * Print the description if debugging
@@ -904,9 +1086,8 @@ class CORE_EXPORT QgsCoordinateReferenceSystem
     void setMapUnits();
 
     //! Helper for getting number of user CRS already in db
-    long getRecordCount();
+    static long getRecordCount();
 
-#if PROJ_VERSION_MAJOR>=6
     bool loadFromAuthCode( const QString &auth, const QString &code );
 
     /**
@@ -922,7 +1103,6 @@ class CORE_EXPORT QgsCoordinateReferenceSystem
      * criteria).
      */
     long matchToUserCrs() const;
-#endif
 
     /**
      * Initialize the CRS object by looking up CRS database in path given in db argument,
@@ -932,24 +1112,16 @@ class CORE_EXPORT QgsCoordinateReferenceSystem
 
     bool createFromWktInternal( const QString &wkt, const QString &description );
 
-#if PROJ_VERSION_MAJOR<6 // not used for proj >= 6.0
-    static bool loadIds( QHash<int, QString> &wkts );
-    static bool loadWkts( QHash<int, QString> &wkts, const char *filename );
-
-    //! Update datum shift definitions from GDAL data. Used by syncDb()
-    static bool syncDatumTransform( const QString &dbPath );
-#endif
-
     QExplicitlySharedDataPointer<QgsCoordinateReferenceSystemPrivate> d;
 
     QString mValidationHint;
 
-#if PROJ_VERSION_MAJOR>=6
+    Qgis::CrsDefinitionFormat mNativeFormat = Qgis::CrsDefinitionFormat::Wkt;
+
     friend class QgsProjContext;
 
     // Only meant to be called by QgsProjContext::~QgsProjContext()
     static void removeFromCacheObjectsBelongingToCurrentThread( PJ_CONTEXT *pj_context );
-#endif
 
     //! Function for CRS validation. May be NULLPTR.
     static CUSTOM_CRS_VALIDATION sCustomSrsValidation;
@@ -972,7 +1144,11 @@ class CORE_EXPORT QgsCoordinateReferenceSystem
     static const QHash< long, QgsCoordinateReferenceSystem > &srIdCache();
 
     friend class TestQgsCoordinateReferenceSystem;
-    friend class QgsPostgresProvider;
+    friend class QgsCoordinateReferenceSystemRegistry;
+    friend bool CORE_EXPORT operator> ( const QgsCoordinateReferenceSystem &c1, const QgsCoordinateReferenceSystem &c2 );
+    friend bool CORE_EXPORT operator< ( const QgsCoordinateReferenceSystem &c1, const QgsCoordinateReferenceSystem &c2 );
+    friend bool CORE_EXPORT operator>= ( const QgsCoordinateReferenceSystem &c1, const QgsCoordinateReferenceSystem &c2 );
+    friend bool CORE_EXPORT operator<= ( const QgsCoordinateReferenceSystem &c1, const QgsCoordinateReferenceSystem &c2 );
 
     bool createFromPostgisSrid( const long id );
 };
@@ -1025,6 +1201,11 @@ inline std::ostream &operator << ( std::ostream &os, const QgsCoordinateReferenc
   // Using streams we need to use local 8 Bit
   return os << mySummary.toLocal8Bit().data() << std::endl;
 }
+
+bool CORE_EXPORT operator> ( const QgsCoordinateReferenceSystem &c1, const QgsCoordinateReferenceSystem &c2 );
+bool CORE_EXPORT operator< ( const QgsCoordinateReferenceSystem &c1, const QgsCoordinateReferenceSystem &c2 );
+bool CORE_EXPORT operator>= ( const QgsCoordinateReferenceSystem &c1, const QgsCoordinateReferenceSystem &c2 );
+bool CORE_EXPORT operator<= ( const QgsCoordinateReferenceSystem &c1, const QgsCoordinateReferenceSystem &c2 );
 #endif
 
 #endif // QGSCOORDINATEREFERENCESYSTEM_H
