@@ -22,6 +22,7 @@
 #include "qgscoordinatereferencesystem.h"
 #include "qgsfeaturesource.h"
 #include "qgsexpressioncontextscopegenerator.h"
+#include "qgscoordinatetransform.h"
 
 #include <QPointer>
 #include <QSet>
@@ -46,7 +47,7 @@ class QgsVectorLayerFeatureIterator;
 
 /**
  * \ingroup core
- * Partial snapshot of vector layer's state (only the members necessary for access to features)
+ * \brief Partial snapshot of vector layer's state (only the members necessary for access to features)
 */
 class CORE_EXPORT QgsVectorLayerFeatureSource : public QgsAbstractFeatureSource
 {
@@ -58,6 +59,11 @@ class CORE_EXPORT QgsVectorLayerFeatureSource : public QgsAbstractFeatureSource
      */
     explicit QgsVectorLayerFeatureSource( const QgsVectorLayer *layer );
 
+    //! QgsVectorLayerFeatureSource cannot be copied
+    QgsVectorLayerFeatureSource( const QgsVectorLayerFeatureSource &other ) = delete;
+    //! QgsVectorLayerFeatureSource cannot be copied
+    QgsVectorLayerFeatureSource &operator==( const QgsVectorLayerFeatureSource &other ) = delete;
+
     ~QgsVectorLayerFeatureSource() override;
 
     QgsFeatureIterator getFeatures( const QgsFeatureRequest &request = QgsFeatureRequest() ) override;
@@ -78,16 +84,23 @@ class CORE_EXPORT QgsVectorLayerFeatureSource : public QgsAbstractFeatureSource
      */
     QgsCoordinateReferenceSystem crs() const;
 
+    /**
+     * Returns the layer id of the source layer.
+     *
+     * \since QGIS 3.4
+     */
+    QString id() const;
+
   protected:
 
-    QgsAbstractFeatureSource *mProviderFeatureSource = nullptr;
-
-    QgsVectorLayerJoinBuffer *mJoinBuffer = nullptr;
-
-    QgsExpressionFieldBuffer *mExpressionFieldBuffer = nullptr;
+    std::unique_ptr< QgsAbstractFeatureSource > mProviderFeatureSource;
+    std::unique_ptr< QgsVectorLayerJoinBuffer > mJoinBuffer;
+    std::unique_ptr< QgsExpressionFieldBuffer > mExpressionFieldBuffer;
 
     QgsFields mFields;
 
+    QString mId;
+
     QgsExpressionContextScope mLayerScope;
 
     bool mHasEditBuffer;
@@ -103,6 +116,11 @@ class CORE_EXPORT QgsVectorLayerFeatureSource : public QgsAbstractFeatureSource
     QgsAttributeList mDeletedAttributeIds;
 
     QgsCoordinateReferenceSystem mCrs;
+
+  private:
+#ifdef SIP_RUN
+    QgsVectorLayerFeatureSource( const QgsVectorLayerFeatureSource &other );
+#endif
 };
 
 /**
@@ -129,22 +147,48 @@ class CORE_EXPORT QgsVectorLayerFeatureIterator : public QgsAbstractFeatureItera
      */
     struct CORE_EXPORT FetchJoinInfo
     {
-      const QgsVectorLayerJoinInfo *joinInfo;//!< Canonical source of information about the join
-      QgsAttributeList attributes;      //!< Attributes to fetch
-      int indexOffset;                  //!< At what position the joined fields start
-      QgsVectorLayer *joinLayer;        //!< Resolved pointer to the joined layer
-      int targetField;                  //!< Index of field (of this layer) that drives the join
-      int joinField;                    //!< Index of field (of the joined layer) must have equal value
+      //! Canonical source of information about the join
+      const QgsVectorLayerJoinInfo *joinInfo;
+      //! Attributes to fetch
+      QgsAttributeList attributes;
+      //! Mapping from original attribute index to the joined layer index
+      QMap<int, int> attributesSourceToDestLayerMap SIP_SKIP;
+      //! At what position the joined fields start
+      int indexOffset;
+
+#ifndef SIP_RUN
+
+      /**
+       * Feature source for join
+       *
+       * \note Not available in Python bindings
+       * \since QGIS 3.20
+       */
+      std::shared_ptr< QgsVectorLayerFeatureSource > joinSource;
+
+      /**
+       * Fields from joined layer.
+       *
+       * \note Not available in Python bindings
+       * \since QGIS 3.20
+       */
+      QgsFields joinLayerFields;
+#endif
+
+      //! Index of field (of this layer) that drives the join
+      int targetField;
+
+      //!< Index of field (of the joined layer) must have equal value
+      int joinField;
 
       void addJoinedAttributesCached( QgsFeature &f, const QVariant &joinValue ) const;
       void addJoinedAttributesDirect( QgsFeature &f, const QVariant &joinValue ) const;
     };
 
-
     bool isValid() const override;
 
   protected:
-    //! fetch next feature, return true on success
+    //! fetch next feature, return TRUE on success
     bool fetchFeature( QgsFeature &feature ) override;
 
     /**
@@ -189,8 +233,9 @@ class CORE_EXPORT QgsVectorLayerFeatureIterator : public QgsAbstractFeatureItera
     /**
      * Adds attributes that don't source from the provider but are added inside QGIS
      * Includes
-     *  - Joined fields
-     *  - Expression fields
+     *
+     * - Joined fields
+     * - Expression fields
      *
      * \param f The feature will be modified
      * \note not available in Python bindings
@@ -223,9 +268,15 @@ class CORE_EXPORT QgsVectorLayerFeatureIterator : public QgsAbstractFeatureItera
     QgsFeatureRequest mChangedFeaturesRequest;
     QgsFeatureIterator mChangedFeaturesIterator;
 
+    // filter bounding box constraint, in SOURCE CRS
     QgsRectangle mFilterRect;
     QgsCoordinateTransform mTransform;
 
+    // distance within constraint reference geometry and distance IN DESTINATION CRS
+    QgsGeometry mDistanceWithinGeom;
+    std::shared_ptr< QgsGeometryEngine > mDistanceWithinEngine;
+    double mDistanceWithin = 0;
+
     // only related to editing
     QSet<QgsFeatureId> mFetchConsidered;
     QgsGeometryMap::ConstIterator mFetchChangedGeomIt;
@@ -235,7 +286,8 @@ class CORE_EXPORT QgsVectorLayerFeatureIterator : public QgsAbstractFeatureItera
 
     /**
      * Information about joins used in the current select() statement.
-      Allows faster mapping of attribute ids compared to mVectorJoins */
+     * Allows faster mapping of attribute ids compared to mVectorJoins.
+    */
     QMap<const QgsVectorLayerJoinInfo *, QgsVectorLayerFeatureIterator::FetchJoinInfo> mFetchJoinInfo;
 
     QMap<int, QgsExpression *> mExpressionFieldInfo;
@@ -247,6 +299,7 @@ class CORE_EXPORT QgsVectorLayerFeatureIterator : public QgsAbstractFeatureItera
     QgsVectorLayerFeatureIterator( const QgsVectorLayerFeatureIterator &rhs );
 #endif
 
+    void createExpressionContext();
     std::unique_ptr<QgsExpressionContext> mExpressionContext;
 
     QgsFeedback *mInterruptionChecker = nullptr;
@@ -258,7 +311,7 @@ class CORE_EXPORT QgsVectorLayerFeatureIterator : public QgsAbstractFeatureItera
     QList< FetchJoinInfo > mOrderedJoinInfoList;
 
     /**
-     * Will always return true. We assume that ordering has been done on provider level already.
+     * Will always return TRUE. We assume that ordering has been done on provider level already.
      *
      */
     bool prepareOrderBy( const QList<QgsFeatureRequest::OrderByClause> &orderBys ) override;
@@ -277,6 +330,8 @@ class CORE_EXPORT QgsVectorLayerFeatureIterator : public QgsAbstractFeatureItera
      * Checks a feature's geometry for validity, if requested in feature request.
      */
     bool checkGeometryValidity( const QgsFeature &feature );
+
+    bool mDelegatedOrderByToProvider = false;
 };
 
 
@@ -284,7 +339,7 @@ class CORE_EXPORT QgsVectorLayerFeatureIterator : public QgsAbstractFeatureItera
 /**
  * \class QgsVectorLayerSelectedFeatureSource
  * \ingroup core
- * QgsFeatureSource subclass for the selected features from a QgsVectorLayer.
+ * \brief QgsFeatureSource subclass for the selected features from a QgsVectorLayer.
  * \since QGIS 3.0
  */
 class CORE_EXPORT QgsVectorLayerSelectedFeatureSource : public QgsFeatureSource, public QgsExpressionContextScopeGenerator
@@ -298,17 +353,26 @@ class CORE_EXPORT QgsVectorLayerSelectedFeatureSource : public QgsFeatureSource,
      */
     QgsVectorLayerSelectedFeatureSource( QgsVectorLayer *layer );
 
+    //! QgsVectorLayerSelectedFeatureSource cannot be copied
+    QgsVectorLayerSelectedFeatureSource( const QgsVectorLayerSelectedFeatureSource &other ) = delete;
+    //! QgsVectorLayerSelectedFeatureSource cannot be copied
+    QgsVectorLayerSelectedFeatureSource &operator==( const QgsVectorLayerSelectedFeatureSource &other ) = delete;
+
     QgsFeatureIterator getFeatures( const QgsFeatureRequest &request = QgsFeatureRequest() ) const override;
     QgsCoordinateReferenceSystem sourceCrs() const override;
     QgsFields fields() const override;
     QgsWkbTypes::Type wkbType() const override;
-    long featureCount() const override;
+    long long featureCount() const override;
     QString sourceName() const override;
     QgsExpressionContextScope *createExpressionContextScope() const override;
-
+    SpatialIndexPresence hasSpatialIndex() const override;
 
   private:
 
+#ifdef SIP_RUN
+    QgsVectorLayerSelectedFeatureSource( const QgsVectorLayerSelectedFeatureSource &other );
+#endif
+
     // ideally this wouldn't be mutable, but QgsVectorLayerFeatureSource has non-const getFeatures()
     mutable QgsVectorLayerFeatureSource mSource;
     QgsFeatureIds mSelectedFeatureIds;
