@@ -21,10 +21,6 @@ __author__ = 'Victor Olaya'
 __date__ = 'August 2012'
 __copyright__ = '(C) 2012, Victor Olaya'
 
-# This will get replaced with a git SHA1 when you do a git archive
-
-__revision__ = '93e761dd2396af21db9bb93e26d7cba680c906ca'
-
 import shutil
 import os
 import sys
@@ -33,29 +29,38 @@ from functools import partial
 from qgis.core import (QgsApplication,
                        QgsProcessingUtils,
                        QgsProcessingModelAlgorithm,
+                       QgsProcessingAlgorithm,
                        QgsDataItemProvider,
                        QgsDataProvider,
                        QgsDataItem,
-                       QgsMapLayer,
-                       QgsMimeDataUtils)
-from qgis.gui import (QgsOptionsWidgetFactory,
+                       QgsMapLayerType,
+                       QgsMimeDataUtils,
+                       QgsSettings)
+from qgis.gui import (QgsGui,
+                      QgsOptionsWidgetFactory,
                       QgsCustomDropHandler)
-from qgis.PyQt.QtCore import Qt, QCoreApplication, QDir, QFileInfo
-from qgis.PyQt.QtWidgets import QMenu, QAction
+from qgis.PyQt.QtCore import QObject, Qt, QItemSelectionModel, QCoreApplication, QDir, QFileInfo, pyqtSlot
+from qgis.PyQt.QtWidgets import QWidget, QMenu, QAction
 from qgis.PyQt.QtGui import QIcon, QKeySequence
 from qgis.utils import iface
 
 from processing.core.Processing import Processing
-from processing.gui.AlgorithmDialog import AlgorithmDialog
 from processing.gui.ProcessingToolbox import ProcessingToolbox
 from processing.gui.HistoryDialog import HistoryDialog
 from processing.gui.ConfigDialog import ConfigOptionsPage
 from processing.gui.ResultsDock import ResultsDock
+from processing.gui.MessageDialog import MessageDialog
+from processing.gui.MessageBarProgress import MessageBarProgress
 from processing.gui.AlgorithmLocatorFilter import (AlgorithmLocatorFilter,
                                                    InPlaceAlgorithmLocatorFilter)
+from processing.gui.Postprocessing import handleAlgorithmResults
+from processing.gui.AlgorithmExecutor import execute, execute_in_place
+from processing.gui.AlgorithmDialog import AlgorithmDialog
+from processing.gui.BatchAlgorithmDialog import BatchAlgorithmDialog
 from processing.modeler.ModelerDialog import ModelerDialog
 from processing.tools.system import tempHelpFolder
-from processing.gui.menus import removeMenus, initializeMenus, createMenus
+from processing.tools import dataobjects
+from processing.gui.menus import removeMenus, initializeMenus, createMenus, createButtons, removeButtons
 from processing.core.ProcessingResults import resultsList
 
 pluginPath = os.path.dirname(__file__)
@@ -89,6 +94,9 @@ class ProcessingDropHandler(QgsCustomDropHandler):
         alg.setProvider(QgsApplication.processingRegistry().providerById('model'))
         dlg = AlgorithmDialog(alg, parent=iface.mainWindow())
         dlg.show()
+        # do NOT remove!!!! if you do then sip forgets the python subclass of AlgorithmDialog and you get a broken
+        # dialog
+        dlg.exec_()
         return True
 
     def customUriProviderKey(self):
@@ -126,7 +134,7 @@ class ProcessingModelItem(QgsDataItem):
         ProcessingDropHandler.runAlg(self.path())
 
     def editModel(self):
-        dlg = ModelerDialog()
+        dlg = ModelerDialog.create()
         dlg.loadModel(self.path())
         dlg.show()
 
@@ -159,10 +167,33 @@ class ProcessingDataItemProvider(QgsDataItemProvider):
         return None
 
 
-class ProcessingPlugin:
+class ProcessingPlugin(QObject):
 
     def __init__(self, iface):
+        super().__init__()
         self.iface = iface
+        self.options_factory = None
+        self.drop_handler = None
+        self.item_provider = None
+        self.locator_filter = None
+        self.edit_features_locator_filter = None
+        self.initialized = False
+        self.initProcessing()
+
+    def initProcessing(self):
+        if not self.initialized:
+            self.initialized = True
+            Processing.initialize()
+
+    def initGui(self):
+        # port old log, ONCE ONLY!
+        settings = QgsSettings()
+        if not settings.value("/Processing/hasPortedOldLog", False, bool):
+            processing_history_provider = QgsGui.historyProviderRegistry().providerById('processing')
+            if processing_history_provider:
+                processing_history_provider.portOldLog()
+                settings.setValue("/Processing/hasPortedOldLog", True)
+
         self.options_factory = ProcessingOptionsFactory()
         self.options_factory.setTitle(self.tr('Processing'))
         iface.registerOptionsWidgetFactory(self.options_factory)
@@ -176,14 +207,14 @@ class ProcessingPlugin:
         iface.currentLayerChanged.connect(lambda _: self.iface.invalidateLocatorResults())
         self.edit_features_locator_filter = InPlaceAlgorithmLocatorFilter()
         iface.registerLocatorFilter(self.edit_features_locator_filter)
-        Processing.initialize()
 
-    def initGui(self):
         self.toolbox = ProcessingToolbox()
         self.iface.addDockWidget(Qt.RightDockWidgetArea, self.toolbox)
         self.toolbox.hide()
         self.toolbox.visibilityChanged.connect(self.toolboxVisibilityChanged)
 
+        self.toolbox.executeWithGui.connect(self.executeAlgorithm)
+
         self.resultsDock = ResultsDock()
         self.iface.addDockWidget(Qt.RightDockWidgetArea, self.resultsDock)
         self.resultsDock.hide()
@@ -205,11 +236,11 @@ class ProcessingPlugin:
 
         self.modelerAction = QAction(
             QgsApplication.getThemeIcon("/processingModel.svg"),
-            QCoreApplication.translate('ProcessingPlugin', 'Graphical &Modeler…'), self.iface.mainWindow())
+            QCoreApplication.translate('ProcessingPlugin', '&Graphical Modeler…'), self.iface.mainWindow())
         self.modelerAction.setObjectName('modelerAction')
         self.modelerAction.triggered.connect(self.openModeler)
         self.iface.registerMainWindowAction(self.modelerAction,
-                                            QKeySequence('Ctrl+Alt+M').toString(QKeySequence.NativeText))
+                                            QKeySequence('Ctrl+Alt+G').toString(QKeySequence.NativeText))
         self.menu.addAction(self.modelerAction)
 
         self.historyAction = QAction(
@@ -225,6 +256,7 @@ class ProcessingPlugin:
         self.resultsAction = QAction(
             QgsApplication.getThemeIcon("/processingResult.svg"),
             self.tr('&Results Viewer'), self.iface.mainWindow())
+        self.resultsAction.setObjectName('resultsViewer')
         self.resultsAction.setCheckable(True)
         self.iface.registerMainWindowAction(self.resultsAction,
                                             QKeySequence('Ctrl+Alt+R').toString(QKeySequence.NativeText))
@@ -262,6 +294,7 @@ class ProcessingPlugin:
 
         initializeMenus()
         createMenus()
+        createButtons()
 
         # In-place editing button state sync
         self.iface.currentLayerChanged.connect(self.sync_in_place_button_state)
@@ -269,6 +302,111 @@ class ProcessingPlugin:
         self.iface.actionToggleEditing().triggered.connect(partial(self.sync_in_place_button_state, None))
         self.sync_in_place_button_state()
 
+        # Sync project models
+        self.projectModelsMenu = None
+        self.projectMenuAction = None
+        self.projectMenuSeparator = None
+
+        self.projectProvider = QgsApplication.instance().processingRegistry().providerById("project")
+        self.projectProvider.algorithmsLoaded.connect(self.updateProjectModelMenu)
+
+    def updateProjectModelMenu(self):
+        """Add projects models to menu"""
+
+        if self.projectMenuAction is None:
+            self.projectModelsMenu = QMenu(self.tr("Models"))
+            self.projectMenuAction = self.iface.projectMenu().insertMenu(self.iface.projectMenu().children()[-1], self.projectModelsMenu)
+            self.projectMenuAction.setParent(self.projectModelsMenu)
+            self.iface.projectMenu().insertSeparator(self.projectMenuAction)
+
+        self.projectModelsMenu.clear()
+
+        for model in self.projectProvider.algorithms():
+            modelSubMenu = self.projectModelsMenu.addMenu(model.name())
+            modelSubMenu.setParent(self.projectModelsMenu)
+            action = QAction(self.tr("Execute…"), modelSubMenu)
+            action.triggered.connect(partial(self.executeAlgorithm, model.id(), self.projectModelsMenu, self.toolbox.in_place_mode))
+            modelSubMenu.addAction(action)
+            if model.flags() & QgsProcessingAlgorithm.FlagSupportsBatch:
+                action = QAction(self.tr("Execute as Batch Process…"), modelSubMenu)
+                modelSubMenu.addAction(action)
+                action.triggered.connect(partial(self.executeAlgorithm, model.id(), self.projectModelsMenu, self.toolbox.in_place_mode, True))
+
+    @pyqtSlot(str, QWidget, bool, bool)
+    def executeAlgorithm(self, alg_id, parent, in_place=False, as_batch=False):
+        """Executes a project model with GUI interaction if needed.
+
+        :param alg_id: algorithm id
+        :type alg_id: string
+        :param parent: parent widget
+        :type parent: QWidget
+        :param in_place: in place flag, defaults to False
+        :type in_place: bool, optional
+        :param as_batch: execute as batch flag, defaults to False
+        :type as_batch: bool, optional
+        """
+
+        config = {}
+        if in_place:
+            config['IN_PLACE'] = True
+
+        alg = QgsApplication.instance().processingRegistry().createAlgorithmById(alg_id, config)
+
+        if alg is not None:
+
+            ok, message = alg.canExecute()
+            if not ok:
+                dlg = MessageDialog()
+                dlg.setTitle(self.tr('Error executing algorithm'))
+                dlg.setMessage(
+                    self.tr('<h3>This algorithm cannot '
+                            'be run :-( </h3>\n{0}').format(message))
+                dlg.exec_()
+                return
+
+            if as_batch:
+                dlg = BatchAlgorithmDialog(alg, iface.mainWindow())
+                dlg.setAttribute(Qt.WA_DeleteOnClose)
+                dlg.show()
+                dlg.exec_()
+            else:
+                in_place_input_parameter_name = 'INPUT'
+                if hasattr(alg, 'inputParameterName'):
+                    in_place_input_parameter_name = alg.inputParameterName()
+
+                if in_place and not [d for d in alg.parameterDefinitions() if d.name() not in (in_place_input_parameter_name, 'OUTPUT')]:
+                    parameters = {}
+                    feedback = MessageBarProgress(algname=alg.displayName())
+                    ok, results = execute_in_place(alg, parameters, feedback=feedback)
+                    if ok:
+                        iface.messageBar().pushSuccess('', self.tr('{algname} completed. %n feature(s) processed.', n=results['__count']).format(algname=alg.displayName()))
+                    feedback.close()
+                    # MessageBarProgress handles errors
+                    return
+
+                if alg.countVisibleParameters() > 0:
+                    dlg = alg.createCustomParametersWidget(parent)
+
+                    if not dlg:
+                        dlg = AlgorithmDialog(alg, in_place, iface.mainWindow())
+                    canvas = iface.mapCanvas()
+                    prevMapTool = canvas.mapTool()
+                    dlg.show()
+                    dlg.exec_()
+                    if canvas.mapTool() != prevMapTool:
+                        try:
+                            canvas.mapTool().reset()
+                        except Exception:
+                            pass
+                        canvas.setMapTool(prevMapTool)
+                else:
+                    feedback = MessageBarProgress(algname=alg.displayName())
+                    context = dataobjects.createContext(feedback)
+                    parameters = {}
+                    ret, results = execute(alg, parameters, context, feedback)
+                    handleAlgorithmResults(alg, context, feedback)
+                    feedback.close()
+
     def sync_in_place_button_state(self, layer=None):
         """Synchronise the button state with layer state"""
 
@@ -277,7 +415,7 @@ class ProcessingPlugin:
 
         old_enabled_state = self.editInPlaceAction.isEnabled()
 
-        new_enabled_state = layer is not None and layer.type() == QgsMapLayer.VectorLayer
+        new_enabled_state = layer is not None and layer.type() == QgsMapLayerType.VectorLayer
         self.editInPlaceAction.setEnabled(new_enabled_state)
 
         if new_enabled_state != old_enabled_state:
@@ -297,11 +435,6 @@ class ProcessingPlugin:
         self.toolbox.deleteLater()
         self.menu.deleteLater()
 
-        # delete temporary output files
-        folder = QgsProcessingUtils.tempFolder()
-        if QDir(folder).exists():
-            shutil.rmtree(folder, True)
-
         # also delete temporary help files
         folder = tempHelpFolder()
         if QDir(folder).exists():
@@ -318,7 +451,16 @@ class ProcessingPlugin:
         self.iface.unregisterCustomDropHandler(self.drop_handler)
         QgsApplication.dataItemProviderRegistry().removeProvider(self.item_provider)
 
+        removeButtons()
         removeMenus()
+
+        if self.projectMenuAction is not None:
+            self.iface.projectMenu().removeAction(self.projectMenuAction)
+            self.projectMenuAction = None
+        if self.projectMenuSeparator is not None:
+            self.iface.projectMenu().removeAction(self.projectMenuSeparator)
+            self.projectMenuSeparator = None
+
         Processing.deinitialize()
 
     def openToolbox(self, show):
@@ -328,7 +470,7 @@ class ProcessingPlugin:
         self.toolboxAction.setChecked(visible)
 
     def openModeler(self):
-        dlg = ModelerDialog()
+        dlg = ModelerDialog.create()
         dlg.update_model.connect(self.updateModel)
         dlg.show()
 
@@ -346,8 +488,8 @@ class ProcessingPlugin:
         dlg = HistoryDialog()
         dlg.exec_()
 
-    def tr(self, message):
-        return QCoreApplication.translate('ProcessingPlugin', message)
+    def tr(self, message, disambiguation=None, n=-1):
+        return QCoreApplication.translate('ProcessingPlugin', message, disambiguation=disambiguation, n=n)
 
     def editSelected(self, enabled):
         self.toolbox.set_in_place_edit_mode(enabled)
